@@ -1,28 +1,44 @@
 use std::collections;
-use std::io::{self, BufRead};
+use std::io::{self, BufRead, BufReader, Write, BufWriter};
+use std::fs::{File, OpenOptions};
 use regex::Regex;
-use structopt::StructOpt;
+use clap::Parser;
 
 
 const PI: f64 = std::f64::consts::PI;
 
-// Command line options
-#[derive(StructOpt)]
-struct Opt {
-    #[structopt(short = "c", long = "centers", help = "Use arc centers in G2/G3")]
+// Command line arguments
+#[derive(Parser)]
+#[clap(version, about, long_about = None,
+       override_usage = "arcfinder [OPTIONS] [INFILE [OUTFILE]]")]
+struct Args {
+    #[clap(short = 'c', long = "centers")]
+    /// Use arc centers in G2/G3
     emit_centers: bool,
 
-    #[structopt(short = "m", long = "matches", default_value = "4")]
+    #[clap(short = 'm', long = "matches", default_value = "4")]
+    /// Minimal number of line segments
     min_match: usize,
 
-    #[structopt(short = "e", long = "error", default_value = "0.01")]
+    #[clap(short = 'e', long = "error", default_value = "0.01")]
+    /// Maximal average movement (in mm)
     rms_limit: f64,
 
-    #[structopt(short = "a", long = "angle", default_value = "40")]
+    #[clap(short = 'a', long = "angle", default_value = "40")]
+    /// Maximal angle (in degrees)
     angle_limit: f64,
 
-    #[structopt(short = "d", long = "deviation", default_value = "0.1")]
-    offset_limit: f64
+    #[clap(short = 'd', long = "deviation", default_value = "0.1")]
+    /// Maximal deviation (in mm)
+    offset_limit: f64,
+
+    #[clap(value_name = "INFILE")]
+    /// Input file
+    infile: Option<String>,
+
+    #[clap(value_name = "OUTFILE", requires = "infile")]
+    /// Output file
+    outfile: Option<String>,
 }
 
 
@@ -45,7 +61,8 @@ impl State {
         self.move_queue.push_back(Move{x: x, y: y, e: e});
     }
 
-    fn process_moves(&mut self, options:&Opt) {
+    // Empty the move queue while outputting processed move commands
+    fn process_moves(&mut self, writer:&mut Box<dyn std::io::Write>, options:&Args) {
         if self.move_queue.is_empty() {
             // No stored moves to process
             return
@@ -80,8 +97,9 @@ impl State {
                     }
                 }
             }
-            
+
             if found_candidate {
+                // We have found an arc. Time to write its G2/G3 command
                 let mut e_sum = 0.0;
                 if self.rel_extrusion {
                     // Calculate the total extrusion
@@ -93,30 +111,31 @@ impl State {
                 }
 
                 if options.emit_centers {
-                    println!("{} X{:5.3} Y{:5.3} I{:5.3} J{:5.3} E{:.5}",
+                    write!(writer, "{} X{:5.3} Y{:5.3} I{:5.3} J{:5.3} E{:.5}\n",
                              if candidate.0 {"G2"} else {"G3"},
                              points[candidate_index].x,
                              points[candidate_index].y,
                              candidate.2.x - self.current_x,
                              candidate.2.y - self.current_y,
-                             e_sum);
+                             e_sum).unwrap();
                 } else {
-                    println!("{} X{:5.3} Y{:5.3} R{:5.3} E{:.5}",
+                    write!(writer, "{} X{:5.3} Y{:5.3} R{:5.3} E{:.5}\n",
                              if candidate.0 {"G2"} else {"G3"},
                              points[candidate_index].x,
                              points[candidate_index].y,
                              candidate.1,
-                             e_sum);
+                             e_sum).unwrap();
                 }
                 
                 self.current_x = points[candidate_index].x;
                 self.current_y = points[candidate_index].y;
                 first = candidate_index+1;
             } else {
+                // We did not find an acceptable arc. Drop first point and try again.
                 self.current_x = points[first].x;
                 self.current_y = points[first].y;
-                println!("G1 X{:5.3} Y{:5.3} E{:.5}",
-                         self.current_x, self.current_y, self.move_queue[first].e);
+                write!(writer, "G1 X{:5.3} Y{:5.3} E{:.5}\n",
+                         self.current_x, self.current_y, self.move_queue[first].e).unwrap();
                 first += 1;
             }
         }
@@ -276,7 +295,7 @@ fn get_angles(a: &Point, b: &Point, c: &Point, points: &[Point], angles: &mut Ve
 }
 
 
-fn find_best_arc(a: &Point, b: &Point, points: &[Point], options:&Opt) -> Option<(bool, f64, Point)> {
+fn find_best_arc(a: &Point, b: &Point, points: &[Point], options:&Args) -> Option<(bool, f64, Point)> {
     let (r, c, rms) = best_arc(a, b, points);
 
     if rms > options.rms_limit {
@@ -319,20 +338,56 @@ fn main() -> Result<(), Box<dyn std::error::Error>> {
     let abs_extrude_pattern = Regex::new(r"^M82\D")?;
     let rel_extrude_pattern = Regex::new(r"^M83\D")?;
 
-    let options = Opt::from_args();
+    let options = Args::parse();
+    let mut temp_file_used = false;
     
-    let stdin = io::stdin();
-    let handle = stdin.lock();
+    // Identity the input stream to read from
+    let reader: Box<dyn BufRead> = match &options.infile {
+        Some(filename) => {
+            let infile = File::open(filename)?;
+            Box::new(BufReader::new(infile))
+        },
+        None => {
+            let stdin = io::stdin();
+            Box::new(BufReader::new(stdin))
+        },
+    };
 
-    for line in handle.lines() {
+    // Identify the output stream to write to
+    let mut writer: Box<dyn Write> = match &options.outfile {
+        Some(filename) => {
+            let outfile = OpenOptions::new().write(true)
+                .create_new(true)
+                .open(filename)?;
+            Box::new(BufWriter::new(outfile))
+        },
+        None => {
+            match &options.infile {
+                Some(filename) => {
+                    // Input file was given, to we should replace input file
+                    let outfile = OpenOptions::new().write(true)
+                        .create_new(true)
+                        .open(filename.to_owned() + ".tmp")?;
+                    temp_file_used = true;
+                    Box::new(BufWriter::new(outfile))
+                },
+                None => {
+                    let stdout = io::stdout();
+                    Box::new(BufWriter::new(stdout))
+                },
+            }
+        },
+    };
+
+    for line in reader.lines() {
         let line = line?;
         match g1_pattern.captures(&line) {
             Some(cap) => state.store_move(cap[1].parse::<f64>()?,
                                           cap[2].parse::<f64>()?,
                                           cap[3].parse::<f64>()?),
             None => {
-                state.process_moves(&options);
-                println!("{}", &line);
+                state.process_moves(&mut writer, &options);
+                write!(writer, "{}\n", &line)?;
 
                 match g0123_x_pattern.captures(&line) {
                     Some(cap) => state.current_x = cap[1].parse::<f64>()?,
@@ -355,7 +410,13 @@ fn main() -> Result<(), Box<dyn std::error::Error>> {
     }
 
     // Empty the queue if something is still there
-    state.process_moves(&options);
+    state.process_moves(&mut writer, &options);
+
+    // Rename temporary file if use, thus overwriting the input
+    if temp_file_used {
+        let basename = options.infile.unwrap().to_string();
+        std::fs::rename(format!("{}.tmp", basename), basename)?;
+    }
     
     Ok(())
 }
